from typing import List, Dict, Any
from ..utils.kvs_api import DEFAULT_TIMEOUT

from ..utils.containers import ClusterConductor
from ..utils.kvs_api import KVSClient
from ..utils.util import Logger


class KVSTestFixture:
    def __init__(self, conductor: ClusterConductor, dir, log: Logger, node_count: int):
        conductor._parent = self
        self.conductor = conductor
        self.dir = dir
        self.node_count = node_count
        self.clients: list[KVSClient] = []
        self.log = log


    def spawn_cluster(self):
        self.log("\n> SPAWN CLUSTER")
        self.conductor.spawn_cluster(node_count=self.node_count)

        for i in range(self.node_count):
            ep = self.conductor.node_external_endpoint(i)
            self.clients.append(KVSClient(ep))

            r = self.clients[i].ping()
            assert r.status_code == 200, f"expected 200 for ping, got {r.status_code}"
            self.log(f"  - node {i} is up: {r.text}")

    def broadcast_view(self, view: List[Dict[str, Any]]):
        self.log(f"\n> SEND VIEW: {view}")
        for i, client in enumerate(self.clients):
            r = client.send_view(view)
            assert r.status_code == 200, (
                f"expected 200 to ack view, got {r.status_code}"
            )
            self.log(f"view sent to node {i}: {r.status_code} {r.text}")

    def rebroadcast_view(self, new_view: Dict[str, List[Dict[str, Any]]]):
        for i, client in enumerate(self.clients):
            r = client.resend_last_view_with_ips_from_new_view(new_view)
            assert (
                r.status_code == 200
            ),  f"expected 200 to ack view, got {r.status_code}"
            self.log(f"view resent to node {i}: {r.status_code} {r.text}")

    def send_view(self, node_id: int, view: Dict[str, List[Dict[str, Any]]]):
        r = self.clients[node_id].send_view(view)
        assert r.status_code == 200, f"expected 200 to ack view, got {r.status_code}"
        self.log(f"view sent to node {node_id}: {r.status_code} {r.text}")

    def destroy_cluster(self):
        self.conductor.dump_all_container_logs(self.dir)
        self.log("\n> DESTROY CLUSTER")
        self.conductor.destroy_cluster()

    def __enter__(self):
        self.spawn_cluster()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.destroy_cluster()


class KVSMultiClient:
    def __init__(self, clients: List[KVSClient], name: str, log: Logger):
        self.clients = clients
        self.metadata = None
        self.name = name
        self.req = 0
        self.log = log

        # internal model of kvs
        self._kvs_model = {}

    def reset_model(self):
        self._kvs_model = {}
        self.metadata = None

    def put(self, node_id: int, key: str, value: str, timeout: float = DEFAULT_TIMEOUT):
        self.log(
            f" {self.name} req_id:{self.req} > {node_id} > kvs.put {key} <- {value}"
        )

        r = self.clients[node_id].put(key, value, self.metadata, timeout=timeout)

        # update model if successful
        if r.status_code // 100 == 2:
            self._kvs_model[key] = value
            self.log(f" {self.name} req_id:{self.req} {r.json()}")
            self.metadata = r.json()["causal-metadata"]

        self.req += 1
        return r

    def get(self, node_id: int, key: str, timeout: float = DEFAULT_TIMEOUT):
        self.log(
            f' {self.name} req_id:{self.req} > {node_id}> kvs.get {
                key
            } request "causal-metadata": {self.metadata}'
        )
        r = self.clients[node_id].get(key, self.metadata, timeout=timeout)

        if r.status_code // 100 == 2:
            self.log(
                f" {self.name} req_id:{self.req} > {node_id}> kvs.get {key} -> {
                    r.json()
                }"
            )
            self.metadata = r.json()["causal-metadata"]
        else:
            self.log(
                f" {self.name} req_id:{self.req} > {node_id}> kvs.get {
                    key
                } -> HTTP ERROR {r.status_code}"
            )

        self.req += 1
        return r

    def get_all(self, node_id: int, timeout: float = DEFAULT_TIMEOUT):
        self.log(
            f' {self.name} req_id:{self.req} > {
                node_id
            }> kvs.get_all request "causal-metadata": {self.metadata}'
        )
        r = self.clients[node_id].get_all(self.metadata, timeout=timeout)
        if r.status_code // 100 == 2:
            self.log(
                f" {self.name} req_id:{self.req} > {node_id}> kvs.get_all -> {r.json()}"
            )
            self.metadata = r.json()["causal-metadata"]
        else:
            self.log(
                f" {self.name} req_id:{self.req} > {
                    node_id
                }> kvs.get_all -> HTTP ERROR {r.status_code}"
            )

        self.req += 1
        return r
